"""Various helpers and utilities."""
from __future__ import annotations

import asyncio
import urllib.error
import urllib.parse
import urllib.request
from contextlib import suppress
from typing import TYPE_CHECKING

from aiohttp import ClientSession

if TYPE_CHECKING:
    from hass_client.models import TokenDetails


def get_websocket_url(url: str) -> str:
    """Extract Websocket URL from Home Assistant URL."""
    if not url or "://" not in url:
        msg = f"{url} is not a valid url"
        raise RuntimeError(msg)
    ws_url = url.replace("http", "ws")
    if not ws_url.endswith("/"):
        ws_url += "/"
    if not ws_url.endswith("/api/"):
        ws_url += "api/"
    return ws_url + "websocket"


def is_supervisor() -> bool:
    """Return if we're running inside the HA Supervisor (e.g. HAOS)."""
    with suppress(urllib.error.URLError):
        return urllib.request.urlopen("ws://supervisor/core/websocket").code == 401  # noqa: S310
    return False


async def async_is_supervisor() -> bool:
    """Return if we're running inside the HA Supervisor (e.g. HAOS), async friendly."""
    loop = asyncio.get_running_loop()
    return await loop.run_in_executor(None, is_supervisor)


def get_auth_url(
    hass_url: str, redirect_uri: str, client_id: str | None = None, state: str | None = None
) -> str:
    """
    Return URL to auth flow.

    The client ID you need to use is the website of your application.
    The redirect url has to be of the same host and port as the client ID.

    If client_id is omitted, it will be derived from the redirect url.

    https://developers.home-assistant.io/docs/auth_api#token
    """
    if not hass_url or "://" not in hass_url:
        msg = f"{hass_url} is not a valid url"
        raise RuntimeError(msg)
    url = hass_url
    if not url.endswith("/"):
        url += "/"
    if not client_id:
        client_id = base_url(redirect_uri)
    enc_client_id = urllib.parse.quote(client_id)
    enc_redirect_uri = urllib.parse.quote(redirect_uri)
    url += f"auth/authorize?client_id={enc_client_id}&redirect_uri={enc_redirect_uri}"
    if state:
        url += f"&state={state}"
    return url


async def get_token(
    hass_url: str, code: str, client_id: str, grant_type: str = "authorization_code"
) -> TokenDetails:
    """
    Return tokens given valid grants.

    The token endpoint returns tokens given valid grants.
    This grant is either an authorization code retrieved from the
    authorize endpoint or a refresh token.

    https://developers.home-assistant.io/docs/auth_api#token
    """
    if not hass_url or "://" not in hass_url:
        msg = f"{hass_url} is not a valid url"
        raise RuntimeError(msg)
    url = hass_url
    if not url.endswith("/"):
        url += "/"
    enc_client_id = urllib.parse.quote(client_id)
    url += "auth/token"
    body = f"grant_type={grant_type}&client_id={enc_client_id}"
    if grant_type == "authorization_code":
        body += f"&code={code}"
    elif grant_type == "refresh_token":
        body += f"&refresh_token={code}"

    async with ClientSession() as session:
        resp = await session.post(
            url, data=body, headers={"Content-Type": "application/x-www-form-urlencoded"}
        )
        result = await resp.json()
        if "error" in result:
            error_msg = result["error"]
            if error_desc := result.get("error_description"):
                error_msg += f": {error_desc}"
            raise RuntimeError(error_msg)
        return result


async def get_long_lived_token(
    hass_url: str,
    access_token: str,
    client_name: str,
    client_icon: str | None = None,
    lifespan: int = 365,
) -> str:
    """Request a Long Lived token with a short lived access token."""
    # prevent circular import
    # pylint: disable-next=import-outside-toplevel
    from .client import HomeAssistantClient

    ws_url = get_websocket_url(hass_url)
    async with HomeAssistantClient(ws_url, access_token) as hass:
        return await hass.send_command(
            "auth/long_lived_access_token",
            client_name=client_name,
            client_icon=client_icon,
            lifespan=lifespan,
        )


def base_url(url: str) -> str:
    """Get the base URL for given url."""
    parsed = urllib.parse.urlparse(url)
    parsed = parsed._replace(path="")
    parsed = parsed._replace(params="")
    parsed = parsed._replace(query="")
    parsed = parsed._replace(fragment="")
    return parsed.geturl()



"""Constants for the Home Assistant Websocket client."""
from typing import Final

# Can be used to specify a catch all when registering state or event listeners.
MATCH_ALL: Final = "*"

# common hass events (for convenience reasons duplicated here)
EVENT_AREA_REGISTRY_UPDATED = "area_registry_updated"
EVENT_DEVICE_REGISTRY_UPDATED = "device_registry_updated"
EVENT_ENTITY_REGISTRY_UPDATED = "entity_registry_updated"
EVENT_STATE_CHANGED = "state_changed"



"""Models used for messages to/from the HA Websocket API."""

from __future__ import annotations

from typing import Any, Final, NotRequired, TypedDict

MESSAGE_TYPE_AUTH: Final[str] = "auth"
MESSAGE_TYPE_AUTH_REQUIRED: Final[str] = "auth"


class AuthRequiredMessage(TypedDict):
    """
    Message received in the auth phase (first message from server after connect).

    https://developers.home-assistant.io/docs/api/websocket#authentication-phase
    """

    type: str  # auth_required
    ha_version: str  # 2021.5.3


class AuthCommandMessage(TypedDict):
    """
    Message sent in the auth phase (first message from client to server).

    https://developers.home-assistant.io/docs/api/websocket#authentication-phase
    """

    type: str  # auth
    access_token: str


class AuthResultMessage(TypedDict):
    """
    Message received as result on the auth command.

    https://developers.home-assistant.io/docs/api/websocket#authentication-phase
    """

    type: str  # auth_ok or auth_invalid
    ha_version: NotRequired[str]
    message: NotRequired[str]  # in case of error


class Message(TypedDict):
    """
    Base Message format for exchanging messages with the server.

    https://developers.home-assistant.io/docs/api/websocket/#command-phase
    """

    type: str
    id: int


class CommandMessage(Message):
    """
    Message when sending a command from client to server.

    https://developers.home-assistant.io/docs/api/websocket/#command-phase
    """


class Context(TypedDict):
    """Context received in an event or result message."""

    id: str
    parent_id: str | None
    user_id: str | None


CommandResultData = dict[str, Any] | list[dict[str, Any]]


class CommandResultMessage(Message):
    """
    Message when receiving the result of a command.

    https://developers.home-assistant.io/docs/api/websocket/#command-phase
    """

    success: bool
    result: CommandResultData


class Event(TypedDict):
    """
    Event object received in Event message.

    https://developers.home-assistant.io/docs/api/websocket/#subscribe-to-events
    """

    event_type: str  # e.g. state_changed
    time_fired: str  # UTC timestamp
    origin: str  # e.g. LOCAL
    context: Context
    data: dict[str, Any]


class Area(TypedDict):
    """Representation of a Home Assistant Area object within the AreaRegistry."""

    aliases: list[str]
    area_id: str
    name: str
    picture: str | None


class Device(TypedDict):
    """Representation of a Home Assistant Device object within the DeviceRegistry."""

    area_id: str | None
    configuration_url: str | None
    config_entries: list[str]
    connections: list[tuple[str, str]]
    disabled_by: str | None
    entry_type: str | None
    hw_version: str | None
    id: str
    identifiers: list[tuple[str, str]]
    manufacturer: str | None
    model: str | None
    name_by_user: str | None
    name: str
    sw_version: str | None
    via_device_id: str | None


class Entity(TypedDict):
    """Representation of a Home Assistant Entity object within the EntityRegistry."""

    area_id: str | None
    config_entry_id: str | None
    device_id: str | None
    disabled_by: str | None
    entity_category: str | None
    entity_id: str
    has_entity_name: bool
    hidden_by: str | None
    icon: str | None
    id: str
    name: str
    options: dict[str, Any]
    original_name: str
    platform: str
    translation_key: str | None
    unique_id: str
    aliases: NotRequired[list[str]]
    capabilities: NotRequired[dict[str, Any]]
    device_class: NotRequired[str | None]
    original_device_class: NotRequired[str | None]
    original_icon: NotRequired[str | None]


class CallServiceResult(TypedDict):
    """Representation of the result received when calling a service."""

    context: Context


class State(TypedDict):
    """Representation of an (entity) State object as received from the api."""

    entity_id: str
    state: str
    attributes: dict[str, Any]
    last_changed: str
    last_updated: str
    context: Context


class UnitSystemDetails(TypedDict):
    """Representation of the Unit system details object as received from the api."""

    length: str  # e.g. km
    accumulated_precipitation: str  # e.g. mm
    mass: str  # e.g. g
    pressure: str  # e.g. Pa
    temperature: str  # e.g. "°C
    volume: str  # e.g. L
    wind_speed: str  # e.g. L


class Config(TypedDict):
    """Representation of the Hass Config object as received from the api."""

    latitude: float
    longitude: float
    elevation: float
    unit_system: UnitSystemDetails
    location_name: str
    time_zone: str
    components: list[str]
    config_dir: str
    whitelist_external_dirs: list[str]
    allowlist_external_dirs: list[str]
    allowlist_external_urls: list[str]
    version: str
    config_source: str
    safe_mode: bool
    state: str
    external_url: str | None
    internal_url: str | None
    currency: str
    country: str
    language: str


class CompressedState(TypedDict):
    """Representation of the compacted State differences object as received from the api."""

    c: NotRequired[str | Context]  # context (id only or full Context obj)
    s: NotRequired[str]  # state
    a: NotRequired[dict[str, Any]]  # state attributes
    lc: NotRequired[float]  # last_changed as timestamp
    lu: NotRequired[float]  # last_updated as timestamp (if differs from last changed)


StateDiff = TypedDict(
    "StateDiff",
    {
        # '+' = in case of attribute additions
        "+": NotRequired[CompressedState],
        # '-' = in case of attribute removals
        "-": NotRequired[CompressedState],
    },
)


class EntityStateEvent(TypedDict):
    """Compact Entity State(event) object received in Event message."""

    # ruff: noqa: E501 pylint: disable=line-too-long
    # {"a":{"light.test":{"s":"on","a":{"min_color_temp_kelvin":2000,"max_color_temp_kelvin":6535,"min_mireds":153,"max_mireds":500,"effect_list":["None","candle","fire"],"supported_color_modes":["color_temp","xy"],"color_mode":"xy","brightness":255,"hs_color":[21.3,78.431],"rgb_color":[255,126,55],"xy_color":[0.5983,0.36],"effect":"None","mode":"normal","dynamics":"dynamic_palette","friendly_name":"Test Light","supported_features":44},"c":"01H0640ES8JCY1NGTNW3V41T5T","lc":1683832716.072648}}}}
    # {"c":{"light.test":{"+":{"lu":1683838800.736819,"c":"01H069T4V03S8D116YDW0GP2WC","a":{"brightness":89}}}}}}

    # NOTE: dict key = entity_id
    a: NotRequired[dict[str, CompressedState]]  # entity added (or first message)
    c: NotRequired[dict[str, StateDiff]]  # entity changed
    r: NotRequired[list[str]]  # entity removed (list of removed entity_ids)


class EntityStateMessage(Message):
    """Compact Entity State(diff) object received in Event message."""

    event: EntityStateEvent


class TokenDetails(TypedDict):
    """Representation of a Token object/message from the token endpoint."""

    access_token: str
    expires_in: int
    refresh_token: str
    token_type: str  # Bearer



"""
Home Assistant Client for python.

Simple wrapper for the Websocket API
provided by Home Assistant that allows for rapid development of apps
connected to Home Assistant.
"""
import asyncio
import logging
import os
import pprint
from collections.abc import Callable
from types import TracebackType
from typing import Any

from aiohttp import (
    ClientSession,
    ClientWebSocketResponse,
    TCPConnector,
    WSMsgType,
    client_exceptions,
)

from .const import MATCH_ALL
from .exceptions import (
    AuthenticationFailed,
    CannotConnect,
    ConnectionFailed,
    FailedCommand,
    InvalidMessage,
    NotConnected,
)
from .models import (
    Area,
    AuthCommandMessage,
    AuthRequiredMessage,
    AuthResultMessage,
    CallServiceResult,
    CommandResultData,
    Config,
    Device,
    Entity,
    EntityStateEvent,
    Event,
    Message,
    State,
)

try:
    import orjson as json

    HAS_ORJSON = True
except ImportError:
    import json

    HAS_ORJSON = False

LOGGER = logging.getLogger(__package__)

EventCallback = Callable[[Event], None]
EntityChangedCallback = Callable[[EntityStateEvent], None]
SubscriptionCallback = Callable[[Message], None]


class HomeAssistantClient:
    """Connection to HomeAssistant (over websockets)."""

    def __init__(
        self,
        websocket_url: str,
        token: str | None,
        aiohttp_session: ClientSession | None = None,
    ) -> None:
        """
        Initialize the connection to HomeAssistant.

        Parameters:
        - websocket_url: full url to the HomeAssistant websocket api or None for supervisor.
        - token: a long lived token or None when using supervisor.
        - aiohttp_session: optionally provide an existing aiohttp session.
        """
        self._websocket_url = websocket_url
        self._token = token
        self._subscriptions: dict[int, tuple[dict, SubscriptionCallback]] = {}
        self._version = None
        self._last_msg_id = 1
        self._loop = asyncio.get_running_loop()
        self._http_session_provided = aiohttp_session is not None
        self._http_session = aiohttp_session
        self._client: ClientWebSocketResponse | None = None
        self._result_futures: dict[str, asyncio.Future] = {}
        self._shutdown_complete_event: asyncio.Event | None = None
        self._msg_id_lock = asyncio.Lock()

    @property
    def connected(self) -> bool:
        """Return if we're currently connected."""
        return self._client is not None and not self._client.closed

    @property
    def version(self) -> str:
        """Return version of connected Home Assistant instance."""
        return self._version

    async def subscribe_events(
        self, cb_func: Callable[[Event], None], event_type: str = MATCH_ALL
    ) -> Callable:
        """
        Subscribe to (all) HA events.

        Parameters:
            - cb_func: callback function or coroutine
            - event_type: Optionally only listen for these event types (defaults to all.)

        Returns: function to remove the listener.
        """

        def handle_message(message: Message):
            if asyncio.iscoroutinefunction(cb_func):
                self._loop.create_task(cb_func(message["event"]))
            else:
                self._loop.call_soon(cb_func, message["event"])

        return await self.subscribe(handle_message, "subscribe_events", event_type=event_type)

    async def subscribe_entities(
        self, cb_func: Callable[[EntityStateEvent], None], entity_ids: list[str]
    ) -> None:
        """
        Subscribe to state_changed events for specific entities only.

        Parameters:
            - cb_func: callback function or coroutine
            - entity_ids: A list of entity_ids to watch.

        Returns: function to remove the listener.

        NOTE: The returned events are a compressed version of the state for performance reasons.
        """

        def handle_message(message: Message):
            if asyncio.iscoroutinefunction(cb_func):
                self._loop.create_task(cb_func(message["event"]))
            else:
                self._loop.call_soon(cb_func, message["event"])

        return await self.subscribe(handle_message, "subscribe_entities", entity_ids=entity_ids)

    async def call_service(
        self,
        domain: str,
        service: str,
        service_data: dict[str, Any] | None = None,
        target: dict[str, Any] | None = None,
    ) -> CallServiceResult:
        """
        Call service on Home Assistant.

        Parameters:
            - domain: Domain of the service to call (e.g. light, switch).
            - service: The service to call  (e.g. turn_on).
            - service_data: Optional dict with parameters (e.g. { brightness: 20 }).
            - target: Optional dict with target parameters (e.g. { device_id: "aabbccddeeffgg" }).
        """
        if not self.connected:
            msg = "Please call connect first."
            raise NotConnected(msg)
        params = {"domain": domain, "service": service}
        if service_data:
            params["service_data"] = service_data
        if target:
            params["target"] = target
        return await self.send_command("call_service", **params)

    async def get_states(self) -> list[State]:
        """Get dump of the current states within Home Assistant."""
        return await self.send_command("get_states")

    async def get_config(self) -> list[Config]:
        """Get dump of the current config in Home Assistant."""
        return await self.send_command("get_states")

    async def get_services(self) -> dict[str, dict[str, Any]]:
        """Get dump of the current services in Home Assistant."""
        return await self.send_command("get_services")

    async def get_area_registry(self) -> list[Area]:
        """Get Area Registry."""
        return await self.send_command("config/area_registry/list")

    async def get_device_registry(self) -> list[Device]:
        """Get Device Registry."""
        return await self.send_command("config/device_registry/list")

    async def get_entity_registry(self) -> list[Entity]:
        """Get Entity Registry."""
        return await self.send_command("config/entity_registry/list")

    async def get_entity_registry_entry(self, entity_id: str) -> Entity:
        """Get single entry from Entity Registry."""
        return await self.send_command("config/entity_registry/get", entity_id=entity_id)

    async def send_command(self, command: str, **kwargs: dict[str, Any]) -> CommandResultData:
        """Send a command to the HA websocket and return response."""
        future: asyncio.Future[CommandResultData] = self._loop.create_future()
        if "message_id" in kwargs:
            message_id = kwargs.pop("message_id")
        else:
            message_id = await self._get_message_id()
        message = {"id": message_id, "type": command, **kwargs}
        self._result_futures[message_id] = future
        await self._send_json_message(message)
        try:
            return await future
        finally:
            self._result_futures.pop(message_id)

    async def send_command_no_wait(
        self, command: str, **kwargs: dict[str, Any]
    ) -> dict[str, Any] | list[dict[str, Any]]:
        """Send a command to the HA websocket without awaiting the response."""
        message_id = await self._get_message_id()
        message = {"id": message_id, "type": command, **kwargs}
        asyncio.create_task(self._send_json_message(message))

    async def subscribe(
        self, cb_func: Callable[[Message], None], command: str, **kwargs: dict[str, Any]
    ) -> Callable:
        """
        Instantiate a subscription for the given command.

        Parameters:
            - cb_func: callback function or coroutine which will be called when a message comes in.
            - command: The command to issue to the server.
            - kwargs: Optionally provider any arguments.

        Returns: function to remove the listener.
        """
        message_base = {"command": command, **kwargs}
        sub = (message_base, cb_func)

        message_id = await self._get_message_id()
        await self.send_command(**message_base, message_id=message_id)
        self._subscriptions[message_id] = sub

        def remove_listener():
            # we need to lookup the key because the subscription id can change due to reconnects
            key = next((x for x, y in self._subscriptions.items() if y == sub), None)
            if not key:
                return
            self._subscriptions.pop(key)
            # try to unsubscribe
            if "subscribe" not in message_base["type"]:
                return
            unsub_command = message_base["type"].replace("subscribe", "unsubscribe")
            asyncio.create_task(self.send_command_no_wait(unsub_command, subscription=key))

        return remove_listener

    async def connect(self) -> None:
        """Connect to the websocket server."""
        if not self._http_session_provided and self._http_session is None:
            self._http_session = ClientSession(
                loop=self._loop, connector=TCPConnector(enable_cleanup_closed=True)
            )
        ws_url = self._websocket_url or "ws://supervisor/core/websocket"
        ws_token = self._token or os.environ.get("HASSIO_TOKEN")
        LOGGER.debug("Connecting to Home Assistant Websocket API on %s", ws_url)
        try:
            self._client = await self._http_session.ws_connect(ws_url, heartbeat=55)
            version_msg: AuthRequiredMessage = await self._client.receive_json()
            self._version = version_msg["ha_version"]
            # send authentication
            auth_command: AuthCommandMessage = {"type": "auth", "access_token": ws_token}
            await self._client.send_json(auth_command)
            auth_result: AuthResultMessage = await self._client.receive_json()
            if auth_result["type"] != "auth_ok":
                raise AuthenticationFailed(auth_result.get("message", "Authentication failed"))
        except (
            client_exceptions.WSServerHandshakeError,
            client_exceptions.ClientError,
        ) as err:
            raise CannotConnect(err) from err

        LOGGER.info(
            "Connected to Home Assistant %s (version %s)",
            self._websocket_url.split("://")[1].split("/")[0],
            self.version,
        )
        # start task to handle incoming messages
        self._loop.create_task(self._process_messages())

    async def disconnect(self) -> None:
        """Disconnect the client."""
        LOGGER.debug("Closing client connection")

        if not self.connected:
            return

        self._shutdown_complete_event = asyncio.Event()
        await self._client.close()

        if not self._http_session_provided and self._http_session:
            await self._http_session.close()
            self._http_session = None
        await self._shutdown_complete_event.wait()

    async def _process_messages(self) -> None:
        """Start listening to the websocket."""
        try:
            while not self._client.closed:
                msg = await self._client.receive()

                if msg.type in (WSMsgType.CLOSE, WSMsgType.CLOSED, WSMsgType.CLOSING):
                    break

                if msg.type == WSMsgType.ERROR:
                    raise ConnectionFailed

                if msg.type != WSMsgType.TEXT:
                    msg = f"Received non-Text message: {msg.type}"
                    raise InvalidMessage(msg)

                try:
                    data = msg.json(loads=json.loads)
                except ValueError as err:
                    msg = "Received invalid JSON."
                    raise InvalidMessage(msg) from err

                if LOGGER.isEnabledFor(logging.DEBUG):
                    LOGGER.debug("Received message:\n%s\n", pprint.pformat(msg))

                self._handle_incoming_message(data)

        finally:
            LOGGER.debug("Listen completed. Cleaning up")

            for future in self._result_futures.values():
                future.cancel()

            if not self._client.closed:
                await self._client.close()

            if self._shutdown_complete_event:
                self._shutdown_complete_event.set()
            else:
                self._on_connection_lost()

    def _handle_incoming_message(self, msg: Message) -> None:
        """Handle incoming message."""
        # command result
        if msg["type"] == "result":
            future = self._result_futures.get(msg["id"])

            if future is None:
                LOGGER.warning("Received result for unknown message with ID: %s", msg["id"])
                return

            if msg["success"]:
                future.set_result(msg["result"])
                return

            future.set_exception(FailedCommand(msg["id"], msg["error"]["message"]))
            return

        # subscription callback
        if msg["id"] in self._subscriptions:
            handler = self._subscriptions[msg["id"]][1]
            if asyncio.iscoroutinefunction(handler):
                self._loop.create_task(handler(msg))
            else:
                self._loop.call_soon(handler, msg)

        # unknown message received, log it
        LOGGER.debug("Received message with unknown type '%s': %s", msg["type"], msg)

    async def _send_json_message(self, message: dict[str, Any]) -> None:
        """Send a message.

        Raises NotConnected if client not connected.
        """
        if not self.connected:
            raise NotConnected

        if LOGGER.isEnabledFor(logging.DEBUG):
            LOGGER.debug("Publishing message:\n%s\n", pprint.pformat(message))

        assert self._client
        assert "id" in message

        if HAS_ORJSON:
            await self._client.send_str(json.dumps(message).decode())
        else:
            await self._client.send_json(message)

    async def __aenter__(self) -> "HomeAssistantClient":
        """Connect to the websocket."""
        await self.connect()
        return self

    async def __aexit__(
        self, exc_type: Exception, exc_value: str, traceback: TracebackType
    ) -> None:
        """Disconnect from the websocket."""
        await self.disconnect()

    def __repr__(self) -> str:
        """Return the representation."""
        prefix = "" if self.connected else "not "
        return f"{type(self).__name__}(ws_server_url={self._websocket_url!r}, {prefix}connected)"

    def _on_connection_lost(self):
        """Call when the connection gets (unexpectedly) lost."""

        async def auto_reconnect():
            """Reconnect the websocket connection when connection lost."""
            attempts = 0
            sleep_time = 2
            while True:
                attempts += 1
                if attempts > 20:
                    sleep_time = 60
                elif sleep_time > 10:
                    sleep_time = 10
                await asyncio.sleep(sleep_time)
                try:
                    await self.connect()
                    # resubscribe all subscriptions
                    subscriptions = list(self._subscriptions.values())
                    self._subscriptions = {}
                    for sub in subscriptions:
                        message_id = await self._get_message_id()
                        await self.send_command(**sub[0], message_id=message_id)
                    return
                except CannotConnect:
                    pass
                if attempts >= 30:
                    LOGGER.warning(
                        "Still could not reconnect after %s attempts, is the server alive ?",
                        attempts,
                    )

        LOGGER.debug("Connection lost, will auto reconnect...")
        self._loop.create_task(auto_reconnect())

    async def _get_message_id(self) -> int:
        """Return a new message id."""
        async with self._msg_id_lock:
            self._last_msg_id = message_id = self._last_msg_id + 1
            return message_id
